/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wAdjoint.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBoundary.h"
#include "wNewton.h"
#include "wPotentialResidual.h"
#include "wFreestreamResidual.h"
#include "wWakeResidual.h"
#include "wKuttaResidual.h"
#include "wLoadFunctional.h"

#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wLinearSolver.h"
#include "wMshDeform.h"
#include "wResults.h"
#include "wMshExport.h"

#include <Eigen/Sparse>

#include <tbb/task_scheduler_init.h>
#include <tbb/tbb.h>
#include <tbb/parallel_do.h>
#include <tbb/spin_mutex.h>

#include <iomanip>

using namespace tbox;
using namespace flow;

#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/**
 * @brief Initialize the adjoint solver
 * @authors Adrien Crovato
 */
Adjoint::Adjoint(std::shared_ptr<Newton> _sol, std::shared_ptr<tbox::MshDeform> _morph, std::shared_ptr<tbox::LinearSolver> _linsol) : sol(_sol), morph(_morph), linsol(_linsol)
{
    // Default parameters
    nthreads = sol->nthreads;
    verbose = sol->verbose;

    // Update element gradients
    sol->pbl->initGradElems();

    // Setup adjoint variables and gradients
    lamClFlo.resize(sol->pbl->msh->nodes.size(), 0.);
    lamCdFlo.resize(sol->pbl->msh->nodes.size(), 0.);
    dClMsh.resize(sol->pbl->msh->nodes.size(), Eigen::Vector3d::Zero());
    dCdMsh.resize(sol->pbl->msh->nodes.size(), Eigen::Vector3d::Zero());
    dClAoa = 0;
    dCdAoa = 0;

    // Setup rows (unknown index)
    this->mapRows();
}

/**
 * @brief Run the adjoint method
 *
 * Solve the adjoint steady transonic Full Potential Equation
 * and compute lift and drag sensitivites
 */
void Adjoint::run()
{
    // Init
    tbb::task_scheduler_init init(nthreads);

    // Display solver parameters
    std::cout << "--- Adjoint solver ---\n"
              << "Number of threads: " << nthreads << "\n"
              << std::endl;

    // Solve adjoint flow equation
    tms["0-AdjFlo"].start();
    this->solve();
    tms["0-AdjFlo"].stop();

    // Compute total gradient of loads with respect to angle of attack
    tms["1-GradAoA"].start();
    this->computeGradientAoa();
    tms["1-GradAoA"].stop();

    // Compute total gradient of loads with respect to mesh nodes (including deformation)
    tms["2-GradMsh"].start();
    this->computeGradientMesh();
    tms["2-GradMsh"].stop();

    // Display timers
    if (verbose > 0)
        std::cout << "Adjoint solver CPU" << std::endl
                  << tms;
    std::cout << std::endl;
}

/**
 * @brief Solve the adjoint steady transonic Full Potential Equation
 */
void Adjoint::solve()
{
    // Build partial gradients of residual wrt to potential
    tms["00-AdjFloJac"].start();
    Eigen::SparseMatrix<double, Eigen::RowMajor> dR(sol->pbl->msh->nodes.size(), sol->pbl->msh->nodes.size());
    Eigen::SparseMatrix<double, Eigen::RowMajor> dR_(sol->pbl->msh->nodes.size(), sol->pbl->msh->nodes.size());
    sol->buildJac(dR_);
    dR = dR_.transpose();
    tms["00-AdjFloJac"].stop();
    tms["01-AdjFloRes"].start();
    // Build partial gradients of loads wrt potential
    Eigen::VectorXd dCl(sol->pbl->msh->nodes.size()), dCd(sol->pbl->msh->nodes.size());
    this->buildGradientLoadsFlow(dCl, dCd);
    tms["01-AdjFloRes"].stop();

    // Solve linear adjoint equations: dR {lambdaCl, lambdaCd} = {dCl, dCd}
    tms["02-AdjFloSol"].start();
    Eigen::Map<Eigen::VectorXd> lamCl_(lamClFlo.data(), lamClFlo.size()), lamCd_(lamCdFlo.data(), lamCdFlo.size());
    linsol->compute(dR, Eigen::Map<Eigen::VectorXd>(dCl.data(), dCl.size()), lamCl_);
    linsol->compute(dR, Eigen::Map<Eigen::VectorXd>(dCd.data(), dCd.size()), lamCd_);
    tms["02-AdjFloSol"].stop();

    // Display
    std::cout << std::setw(15) << std::left << "Flow adjoint"
              << std::setw(15) << std::right << "Res[lambdaCl]"
              << std::setw(15) << std::right << "Res[lambdaCd]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << log10((dR * lamCl_ - dCl).norm())
              << std::setw(15) << std::right << log10((dR * lamCd_ - dCd).norm()) << std::endl;
}

/**
 * @brief Compute total gradients of the loads with respect to angle of attack
 * @todo one gradient for each surface node
 */
void Adjoint::computeGradientAoa()
{
    // Compute partial gradients with respect to AoA
    tms["10-GradAoARes"].start();
    Eigen::VectorXd dR = Eigen::VectorXd::Zero(sol->pbl->msh->nodes.size());
    this->buildGradientResidualAoa(dR);
    tms["10-GradAoARes"].stop();
    tms["11-GradAoALoad"].start();
    double dL = 0, dD = 0;
    this->buildGradientLoadsAoa(dL, dD);
    tms["11-GradAoALoad"].stop();
    // Compute total gradients
    dClAoa = (dL - Eigen::Map<Eigen::VectorXd>(lamClFlo.data(), lamClFlo.size()).transpose() * dR);
    dCdAoa = (dD - Eigen::Map<Eigen::VectorXd>(lamCdFlo.data(), lamCdFlo.size()).transpose() * dR);

    // Display
    std::cout << std::setw(15) << std::left << "AoA gradients"
              << std::setw(15) << std::right << "dCl_dAoA"
              << std::setw(15) << std::right << "dCd_dAoA" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << dClAoa
              << std::setw(15) << std::right << dCdAoa << std::endl;
}

/**
 * @brief Compute total gradients of the loads with respect to mesh nodes
 */
void Adjoint::computeGradientMesh()
{
    // Build partial gradients of flow residual wrt mesh
    tms["20-GradMshRes"].start();
    Eigen::SparseMatrix<double, Eigen::RowMajor> dR(sol->pbl->msh->nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    this->buildGradientResidualMesh(dR);
    tms["20-GradMshRes"].stop();
    // Build partial gradients of loads wrt mesh
    tms["21-GradMshLoad"].start();
    Eigen::RowVectorXd dCl = Eigen::RowVectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size()), dCd = Eigen::RowVectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size());
    this->buildGradientLoadsMesh(dCl, dCd);
    tms["21-GradMshLoad"].stop();

    // Build partial gradients of mesh deformation residual wrt mesh
    tms["22-GradMshAdj"].start();
    Eigen::SparseMatrix<double, Eigen::RowMajor> Kmesh(sol->pbl->nDim * sol->pbl->msh->nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    Eigen::SparseMatrix<double, Eigen::RowMajor> Kmesh_(sol->pbl->nDim * sol->pbl->msh->nodes.size(), sol->pbl->nDim * sol->pbl->msh->nodes.size());
    morph->build(Kmesh_);
    Kmesh = Kmesh_.transpose();
    // Build RHS of mesh deformation adjoint equations: {dCl, dCd} = {dCl, dCd}_x - {lambdaCl, lambdaCd}_flow * dR_x
    Eigen::VectorXd dClXyz = (dCl - Eigen::Map<Eigen::VectorXd>(lamClFlo.data(), lamClFlo.size()).transpose() * dR).transpose();
    Eigen::VectorXd dCdXyz = (dCd - Eigen::Map<Eigen::VectorXd>(lamCdFlo.data(), lamCdFlo.size()).transpose() * dR).transpose();
    tms["22-GradMshAdj"].stop();
    // Solve mesh deformation adjoint equations: K {lambdaCl, lambdaCd}_mesh = {dCl, dCd}
    tms["23-GradMshSol"].start();
    Eigen::VectorXd lamClMsh = Eigen::VectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size());
    Eigen::VectorXd lamCdMsh = Eigen::VectorXd::Zero(sol->pbl->nDim * sol->pbl->msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> lamClMsh_(lamClMsh.data(), lamClMsh.size()), lamCdMsh_(lamCdMsh.data(), lamCdMsh.size());
    linsol->compute(Kmesh, Eigen::Map<Eigen::VectorXd>(dClXyz.data(), dClXyz.size()), lamClMsh_);
    linsol->compute(Kmesh, Eigen::Map<Eigen::VectorXd>(dCdXyz.data(), dCdXyz.size()), lamCdMsh_);
    tms["23-GradMshSol"].stop();
    // Store solution
    for (auto n : sol->pbl->msh->nodes)
    {
        for (int m = 0; m < sol->pbl->nDim; m++)
        {
            dClMsh[n->row](m) = lamClMsh(sol->pbl->nDim * (n->row) + m);
            dCdMsh[n->row](m) = lamCdMsh(sol->pbl->nDim * (n->row) + m);
        }
    }

    // Display
    std::cout << std::setw(15) << std::left << "Mesh adjoint"
              << std::setw(15) << std::right << "Res[lambdaCl]"
              << std::setw(15) << std::right << "Res[lambdaCd]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << log10((Kmesh * lamClMsh - dClXyz).norm())
              << std::setw(15) << std::right << log10((Kmesh * lamCdMsh - dCdXyz).norm()) << std::endl;
    std::cout << std::setw(15) << std::left << "Mesh gradients"
              << std::setw(15) << std::right << "Norm[dCl_dX]"
              << std::setw(15) << std::right << "Norm[dCd_dX]" << std::endl;
    std::cout << std::setw(15) << std::left << ""
              << std::setw(15) << std::right << lamClMsh.norm()
              << std::setw(15) << std::right << lamCdMsh.norm() << std::endl;
}

/**
 * @brief Build the gradients of the lift and drag with respect to the flow variables
 * @todo one gradient for each surface node
 */
void Adjoint::buildGradientLoadsFlow(Eigen::VectorXd &dL, Eigen::VectorXd &dD)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Reset and compute gradient of lift and drag
    dL.setZero();
    dD.setZero();
    for (auto sur : sol->pbl->bnds)
    {
        tbb::parallel_do(sur->groups[0]->tag->elems.begin(), sur->groups[0]->tag->elems.end(), [&](Element *e) {
            // Associated volume element
            Element *eV = sur->svMap.at(e);
            // Build
            Eigen::MatrixXd Ae = LoadFunctional::buildGradientFlow(*e, *eV, sol->phi, *sol->pbl->medium->cP);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < e->nodes.size(); ++i)
            {
                Eigen::RowVectorXd AeL = sol->pbl->dirL.eval().transpose() * Ae.block(i * 3, 0, 3, eV->nodes.size());
                Eigen::RowVectorXd AeD = sol->pbl->dirD.eval().transpose() * Ae.block(i * 3, 0, 3, eV->nodes.size());
                for (size_t j = 0; j < eV->nodes.size(); ++j)
                {
                    Node *nodj = eV->nodes[j];
                    dL(nodj->row) += AeL(j);
                    dD(nodj->row) += AeD(j);
                }
            }
        });
    }
}

/**
 * @brief Compute gradients of the residual with respect to angle of attack
 */
void Adjoint::buildGradientResidualAoa(Eigen::VectorXd &dR)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Reset gradient
    dR.setZero();

    // Freestream contribution
    for (auto nBC : sol->pbl->fBCs)
    {
        tbb::parallel_do(nBC->tag->elems.begin(), nBC->tag->elems.end(), [&](Element *e) {
            // Build elementary flux vector
            Eigen::VectorXd be = FreestreamResidual::buildGradientAoa(*e, sol->phi, *nBC);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                dR(sol->rows[nodi->row]) += be(ii);
            }
        });
    }
}

/**
 * @brief Compute gradients of the lift and drag with respect to angle of attack
 * @todo one gradient for each surface node
 */
void Adjoint::buildGradientLoadsAoa(double &dL, double &dD)
{
    // Compute integrated aerodynamic load coefficients (normalized by freestream dynamic pressure and reference area)
    Eigen::Vector3d Cf(0, 0, 0);
    for (auto bnd : sol->pbl->bnds)
        for (size_t i = 0; i < bnd->nodes.size(); ++i)
            Cf += bnd->nLoads[i];
    // Rotate to gradient of flow direction
    dD = Cf.dot(sol->pbl->dirD.evalGrad());
    dL = Cf.dot(sol->pbl->dirL.evalGrad());
}

/**
 * @brief Compute gradients of the residual with respect to mesh nodes
 */
void Adjoint::buildGradientResidualMesh(Eigen::SparseMatrix<double, Eigen::RowMajor> &dR)
{
    // Multithread
    tbb::spin_mutex mutex;

    // List of triplets
    std::deque<Eigen::Triplet<double>> T;

    // Volume
    auto fluid = sol->pbl->medium;
    tbb::parallel_do(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;
        //Upwind element
        Element *eU = p.second[0];
        // Build elementary matrices
        Eigen::MatrixXd Ae1, Ae2;
        std::tie(Ae1, Ae2) = PotentialResidual::buildGradientMesh(*e, *eU, sol->phi, *fluid, sol->pbl->nDim, sol->muCv, sol->mCOv);

        // Assembly (subsonic)
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            size_t rowi = e->nodes[i]->row;
            for (size_t j = 0; j < e->nodes.size(); ++j)
            {
                for (int n = 0; n < sol->pbl->nDim; ++n)
                {
                    for (size_t jj = 0; jj < arows[e->nodes[j]->row].size(); ++jj)
                    {
                        size_t rowj = sol->pbl->nDim * (arows[e->nodes[j]->row][jj]) + n;
                        T.push_back(Eigen::Triplet<double>(sol->rows[rowi], rowj, Ae1(i, sol->pbl->nDim * j + n)));
                    }
                }
            }
        }
        // Assembly (supersonic)
        double mach = fluid->mach->eval(*e, sol->phi, 0);
        if (mach > sol->mCOv)
        {
            for (size_t i = 0; i < e->nodes.size(); ++i)
            {
                size_t rowi = e->nodes[i]->row;
                for (size_t j = 0; j < eU->nodes.size(); ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[eU->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[eU->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(sol->rows[rowi], rowj, Ae2(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
            }
        }
    });
    // Farfield B.C. are not taken into account because they have no influence on the solution

    // Wake
    for (auto wake : sol->pbl->wBCs)
    {
        tbb::parallel_do(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            // Build elementary matrices
            Eigen::MatrixXd Aeu, Ael, Aes;
            std::tie(Aeu, Ael, Aes) = WakeResidual::buildGradientMesh(*we, sol->phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < we->nRow; ++i)
            {
                size_t rowi = we->wkN[i].second->row;
                for (size_t j = 0; j < we->nColUp; ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[we->volUpE->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[we->volUpE->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(rowi, rowj, Aeu(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
                for (size_t j = 0; j < we->nColLw; ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[we->volLwE->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[we->volLwE->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(rowi, rowj, -Ael(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
                for (size_t j = 0; j < we->surUpE->nodes.size(); ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[we->surUpE->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[we->surUpE->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(rowi, rowj, Aes(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
            }
        });
    }
    // Kutta
    for (auto kutta : sol->pbl->kSCs)
    {
        tbb::parallel_do(kutta->kEle.begin(), kutta->kEle.end(), [&](KuttaElement *ke) {
            // Build elementary matrices
            Eigen::MatrixXd Ae, Aes;
            std::tie(Ae, Aes) = KuttaResidual::buildGradientMesh(*ke, sol->phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < ke->nRow; ++i)
            {
                size_t rowi = ke->teN[i].second->row;
                for (size_t j = 0; j < ke->nCol; ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[ke->volE->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[ke->volE->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(rowi, rowj, Ae(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
                for (size_t j = 0; j < ke->surE->nodes.size(); ++j)
                {
                    for (int n = 0; n < sol->pbl->nDim; ++n)
                    {
                        for (size_t jj = 0; jj < arows[ke->surE->nodes[j]->row].size(); ++jj)
                        {
                            size_t rowj = sol->pbl->nDim * (arows[ke->surE->nodes[j]->row][jj]) + n;
                            T.push_back(Eigen::Triplet<double>(rowi, rowj, Aes(i, sol->pbl->nDim * j + n)));
                        }
                    }
                }
            }
        });
    }
    // Assemble matrix
    dR.setFromTriplets(T.begin(), T.end());
    dR.prune(0.);
    dR.makeCompressed();
}

/**
 * @brief Compute gradients of the lift and drag with respect to mesh nodes
 * @todo one gradient for each surface node
 */
void Adjoint::buildGradientLoadsMesh(Eigen::RowVectorXd &dL, Eigen::RowVectorXd &dD)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Reset and compute gradient of lift and drag
    dL.setZero();
    dD.setZero();
    for (auto bnd : sol->pbl->bnds)
    {
        tbb::parallel_do(bnd->groups[0]->tag->elems.begin(), bnd->groups[0]->tag->elems.end(), [&](Element *e) {
            // Volume element
            Element *eV = bnd->svMap.at(e);
            // Build
            Eigen::MatrixXd Aes, Aev;
            std::tie(Aes, Aev) = LoadFunctional::buildGradientMesh(*e, *eV, sol->phi, *sol->pbl->medium->cP, sol->pbl->nDim);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t i = 0; i < e->nodes.size(); ++i)
            {
                // rotate to flow direction
                Eigen::RowVectorXd AesL = sol->pbl->dirL.eval().transpose() * Aes.block(i * 3, 0, 3, sol->pbl->nDim * e->nodes.size());
                Eigen::RowVectorXd AesD = sol->pbl->dirD.eval().transpose() * Aes.block(i * 3, 0, 3, sol->pbl->nDim * e->nodes.size());
                Eigen::RowVectorXd AevL = sol->pbl->dirL.eval().transpose() * Aev.block(i * 3, 0, 3, sol->pbl->nDim * eV->nodes.size());
                Eigen::RowVectorXd AevD = sol->pbl->dirD.eval().transpose() * Aev.block(i * 3, 0, 3, sol->pbl->nDim * eV->nodes.size());
                for (size_t m = 0; m < sol->pbl->nDim; ++m)
                {
                    // surface
                    for (size_t j = 0; j < e->nodes.size(); ++j)
                    {
                        size_t rowj = e->nodes[j]->row;
                        for (size_t jj = 0; jj < arows[rowj].size(); ++jj)
                        {
                            dL.col((arows[rowj][jj]) * sol->pbl->nDim + m) += AesL.col(j * sol->pbl->nDim + m);
                            dD.col((arows[rowj][jj]) * sol->pbl->nDim + m) += AesD.col(j * sol->pbl->nDim + m);
                        }
                    }
                    // volume
                    for (size_t j = 0; j < eV->nodes.size(); ++j)
                    {
                        size_t rowj = eV->nodes[j]->row;
                        for (size_t jj = 0; jj < arows[rowj].size(); ++jj)
                        {
                            dL.col((arows[rowj][jj]) * sol->pbl->nDim + m) += AevL.col(j * sol->pbl->nDim + m);
                            dD.col((arows[rowj][jj]) * sol->pbl->nDim + m) += AevD.col(j * sol->pbl->nDim + m);
                        }
                    }
                }
            }
        });
    }
}

/**
 * @brief Create the rows vector for the adjoint mesh
 * @todo duplicated from wake
 */
void Adjoint::mapRows()
{
    // Collect each row associated to each node
    arows.resize(sol->pbl->msh->nodes.size());
    for (size_t i = 0; i < sol->pbl->msh->nodes.size(); ++i)
        arows[i].push_back(sol->pbl->msh->nodes[i]->row);

    // Add duplicated row on the wake
    for (auto wake : sol->pbl->wBCs)
    {
        // Collect nodes on the wake
        std::vector<Node *> nodesUp, nodesLw;
        for (auto e : wake->groups[0]->tag->elems)
            for (auto n : e->nodes)
                nodesUp.push_back(n);
        for (auto e : wake->groups[1]->tag->elems)
            for (auto n : e->nodes)
                nodesLw.push_back(n);
        std::sort(nodesUp.begin(), nodesUp.end());
        auto it = std::unique(nodesUp.begin(), nodesUp.end());
        nodesUp.resize(std::distance(nodesUp.begin(), it));
        std::sort(nodesLw.begin(), nodesLw.end());
        it = std::unique(nodesLw.begin(), nodesLw.end());
        nodesLw.resize(std::distance(nodesLw.begin(), it));
        // Create map between upper and lower sides, but only retain different nodes
        std::map<Node *, Node *> nodMap;
        for (auto nUp : nodesUp)
        {
            for (auto nLw : nodesLw)
            {
                if ((nUp->pos - nLw->pos).norm() <= 1e-14 && nUp != nLw)
                {
                    nodMap[nUp] = nLw;
                    break;
                }
            }
        }
        // Create rows
        for (auto p : nodMap)
        {
            arows[p.first->row].push_back(p.second->row);
            arows[p.second->row].push_back(p.first->row);
        }
    }
}

/**
 * @brief Write the results
 */
void Adjoint::save(MshExport *mshWriter, int n)
{
    // Write files
    std::cout << "Saving files... " << std::endl;
    // setup results
    Results results;
    results.scalars_at_nodes["lambdaClPhi"] = &lamClFlo;
    results.scalars_at_nodes["lambdaCdPhi"] = &lamCdFlo;
    results.vectors_at_nodes["gradClMsh"] = &dClMsh;
    results.vectors_at_nodes["gradCdMsh"] = &dCdMsh;
    // save (all mesh and boundary surface)
    if (n > 0)
    {
        mshWriter->save(sol->pbl->msh->name + "_adjoint_" + std::to_string(n), results);
        for (auto sur : sol->pbl->bnds)
            sur->save(sur->groups[0]->tag->name + "adjoint_" + std::to_string(n), results);
    }
    else
    {
        mshWriter->save(sol->pbl->msh->name + "_adjoint", results);
        for (auto sur : sol->pbl->bnds)
            sur->save(sur->groups[0]->tag->name + "_adjoint", results);
    }
}

void Adjoint::write(std::ostream &out) const
{
    out << "flow::Adjoint"
        << "\n";
}
